/*-------------------------------------------------------------------------
	Copyright (C) 2008-2009 American Megatrends Inc.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-------------------------------------------------------------------------*/
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <linux/i2c.h>
#include <linux/i2c-dev.h>
#include "i2c-data.h"
#include "i2c-hardware.h"
#include "gpio_vsc.h"
#include "gpio_info.h"
#include <linux/kobject.h>
#include <linux/string.h>
#include <linux/sysfs.h>
#include <linux/timer.h>
#include <asm-mips/mach-vsc452/vsc452.h>

#define VERSION_STR 	"7.1"

#define I2C_ID			1
#define IRQ_I2C			7

#define I2C_REMAP_FAILS			0x0A
#define I2C_MEM_CHECK_REG_FAILS	0xBB
#define I2C_REQ_MEM_REG_FAILS	0xCC
#define I2C_REGIS_IRQ_FAILS		0xDD
#define SET_IIC_OPTION			2


#define MLER_LOC_REL_SHORT_TMO		0x01
#define MLER_LOC_REL_LONG_TMO		0x02
#define MLER_RM_GRT_TMO				0x04
#define MLER_RM_REL_TMO				0x08
#define MLER_LOC_SER_EXCH_ERR		0x10

#define MRER_RM_REL_SHORT_TMO		0x01
#define MRER_RM_REL_LONG_TMO		0x02
#define MRER_LOC_GRT_TMO			0x04
#define MRER_LOC_REL_TMO			0x08
#define MRER_RM_SER_EXCH_ERR		0x10


#ifdef ARBITRATION_SUPPORT
struct timer_list i2c_timer;
#endif

typedef struct
{
	int segnum;
	unsigned short error_bit;
} segment_err_T;
 

int i2c_dev_id = 1;
/*EMW
int i2c_node[BUS_COUNT] =
{
#if (BUS_COUNT >= 1)
	I2C_NODE_0, 
#endif
#if (BUS_COUNT >= 2)
	I2C_NODE_1, 
#endif
#if (BUS_COUNT >= 3)
	I2C_NODE_2, 
#endif
#if (BUS_COUNT >= 4)
	I2C_NODE_3, 
#endif
#if (BUS_COUNT >= 5)
	I2C_NODE_4, 
 #endif
 #if (BUS_COUNT >= 6)
 	I2C_NODE_5
 #endif
 };
 */
 
static struct i2c_adapter vsc_ops[BUS_COUNT];
extern struct i2c_vsc_data vsc_data_ptr[BUS_COUNT];


 
extern int write_init_st(int option,int value);

void *i2c_remap_addr;
static int bus_count=BUS_COUNT;
static short init_iic_status=0;


extern int vsc_get_recovery_info(int bus, bus_recovery_info_T* info);
extern int vsc_set_recovery_info(int bus, bus_recovery_info_T* info);

gpio_dev_t *g_gpio_dev=NULL;
extern void * get_gpio_dev(void);




extern unsigned char  g_mux_chan[BUS_COUNT];

/* /sys/kernel Interface variable */
static struct kobject *i2c_kobj; 				   // i2c dir (/sys/kernel/i2c)
static struct kobject *i2c_ports[BUS_COUNT];	   // Bus index ( /sys/kernle/i2c/x) 


extern i2c_bus_status_t i2c_bus_status_ptr[BUS_COUNT];
extern sharedbus_stats_t g_sharedbus_stats[SEGMENT_COUNT];	








/*********************************************************************************************************************
*
*               /sys/kernel/i2c  Interface 
*
***********************************************************************************************************************/

/*
* Helper functions for /sys Interface
*/  


static int 
bus_name_to_number(struct kobject *bus_num) {
	int retval = -EINVAL;

	if (bus_num) {
		if (bus_num->name) {
			retval = *(bus_num->name) - '0';
			if ((retval < 0) && (retval > BUS_COUNT)) {
				retval = -EINVAL;
			}
		}
	}

    return retval;
}


static ssize_t
print_bus_status(int bus_num, char *buf, size_t *len) {
	

	*len += sprintf(buf + *len, "Bus %d\n",bus_num);
	*len += sprintf(buf + *len, "Master bytes sent                            %d\n", i2c_bus_status_ptr[bus_num].TX_index);
	*len += sprintf(buf + *len, "Master bytes received                        %d\n", i2c_bus_status_ptr[bus_num].MasterRX_index);
	*len += sprintf(buf + *len, "Interrupt count                              %d\n", i2c_bus_status_ptr[bus_num].int_cnt);
	*len += sprintf(buf + *len, "Recovery count                               %d\n", i2c_bus_status_ptr[bus_num].bus_recovery_cnt);
	*len += sprintf(buf + *len, "Busy count                                   %d\n", i2c_bus_status_ptr[bus_num].bus_busy_cnt);
	*len += sprintf(buf + *len, "Consequtive busy count                       %d\n", i2c_bus_status_ptr[bus_num].bus_conseq_busy_cnt);
	*len += sprintf(buf + *len, "Failure count                                %d\n\n", i2c_bus_status_ptr[bus_num].bus_fail_cnt);
	return *len;
}


static ssize_t
print_segment_status(int seg_num, char *buf, int *len) {


	*len += sprintf(buf + *len, "Shared Bus Segment %d\n", seg_num);	
	*len += sprintf(buf + *len, "Local  Error - Local release short timeout   %d\n", g_sharedbus_stats[seg_num].l_rel_short_timeout);
	*len += sprintf(buf + *len, "Local  Error - Local release long timeout    %d\n", g_sharedbus_stats[seg_num].l_rel_long_timeout);
	*len += sprintf(buf + *len, "Local  Error - Remote grant timeout          %d\n", g_sharedbus_stats[seg_num].l_rm_grant_timeout);
	*len += sprintf(buf + *len, "Local  Error - Remote release	timeout       %d\n", g_sharedbus_stats[seg_num].l_rm_rel_timeout);
	*len += sprintf(buf + *len, "Local  Error - Local serial exchange error	  %d\n", g_sharedbus_stats[seg_num].l_serial_exchange_error);
	*len += sprintf(buf + *len, "Remote Error - Remote release short timeout  %d\n", g_sharedbus_stats[seg_num].r_rel_short_timeout);
	*len += sprintf(buf + *len, "Remote Error - Remote release long timeout   %d\n", g_sharedbus_stats[seg_num].r_rel_long_timeout);
	*len += sprintf(buf + *len, "Remote Error - Local grant timeout           %d\n", g_sharedbus_stats[seg_num].r_loc_grant_timeout);
	*len += sprintf(buf + *len, "Remote Error - Local release timeout         %d\n", g_sharedbus_stats[seg_num].r_loc_rel_timeout);
	*len += sprintf(buf + *len, "Remote Error - Remote serial exchange error  %d\n\n", g_sharedbus_stats[seg_num].r_serial_exchange_error);

	return *len;
}
void
clear_segment_status(int seg_num) {

	g_sharedbus_stats[seg_num].l_rel_short_timeout = 0;
	g_sharedbus_stats[seg_num].l_rel_long_timeout = 0;
	g_sharedbus_stats[seg_num].l_rm_grant_timeout = 0;
	g_sharedbus_stats[seg_num].l_rm_rel_timeout = 0;
	g_sharedbus_stats[seg_num].l_serial_exchange_error = 0;
	g_sharedbus_stats[seg_num].r_rel_short_timeout = 0;
	g_sharedbus_stats[seg_num].r_rel_long_timeout = 0;
	g_sharedbus_stats[seg_num].r_loc_grant_timeout = 0;
	g_sharedbus_stats[seg_num].r_loc_rel_timeout = 0;
	g_sharedbus_stats[seg_num].r_serial_exchange_error = 0;

	return;
}
static ssize_t
print_dev_status(int bus_num,unsigned char dev_addr,char *buf, int *len) {

	//unsigned char i=0;
	char mode[20];
	char muxslave[20];
	char muxchan[20];
	
	if (i2c_bus_status_ptr[bus_num].dev_stat[dev_addr].present == VALID) {
			
		if (i2c_bus_status_ptr[bus_num].dev_stat[dev_addr].mode == 0)
			sprintf (mode, "idle");
		else if (i2c_bus_status_ptr[bus_num].dev_stat[dev_addr].mode == 1)
			sprintf (mode, "master transmit");
		else if (i2c_bus_status_ptr[bus_num].dev_stat[dev_addr].mode == 2)
			sprintf (mode, "master  receive");
		else if ( i2c_bus_status_ptr[bus_num].dev_stat[dev_addr].mode == 6)
			sprintf (mode, "master tran_rec");
		else
			sprintf (mode, "invalid");


		if (i2c_bus_status_ptr[bus_num].dev_stat[dev_addr].mux_slave == 0)
			sprintf(muxslave,"na");
		else 
			sprintf(muxslave, "0x%x",i2c_bus_status_ptr[bus_num].dev_stat[dev_addr].mux_slave);

		/*** for all the devices without mux before them (all the devices not in bus 2 and bus 3)*/
		if ((i2c_bus_status_ptr[bus_num].dev_stat[dev_addr].mux_chan == 0)&& (bus_num !=2) &&(bus_num !=3))
			sprintf(muxchan,"na");
		else
			sprintf(muxchan,"0x%x",i2c_bus_status_ptr[bus_num].dev_stat[dev_addr].mux_chan);
			
		*len += sprintf(buf + *len, "Device (bus: %d; mux slave: %s, chan: %s; slave: 0x%x)\n", 
			bus_num,muxslave,muxchan,i2c_bus_status_ptr[bus_num].dev_stat[dev_addr].dev_slave_addr << 1);	
			
		*len += sprintf(buf + *len, "Mode                                         %s\n", mode);	
		*len += sprintf(buf + *len, "Master bytes sent                            %d\n", i2c_bus_status_ptr[bus_num].dev_stat[dev_addr].xmit_cnt);
		*len += sprintf(buf + *len, "Master bytes received                        %d\n", i2c_bus_status_ptr[bus_num].dev_stat[dev_addr].recv_cnt);
		*len += sprintf(buf + *len, "Recovery count                               %d\n", i2c_bus_status_ptr[bus_num].dev_stat[dev_addr].recovery_cnt);
		*len += sprintf(buf + *len, "Busy counts                                  %d\n", i2c_bus_status_ptr[bus_num].dev_stat[dev_addr].busy_cnt);
		*len += sprintf(buf + *len, "Consequtive Failure count                    %d\n", i2c_bus_status_ptr[bus_num].dev_stat[dev_addr].conseq_fail_cnt);
		*len += sprintf(buf + *len, "Failure count                                %d\n", i2c_bus_status_ptr[bus_num].dev_stat[dev_addr].failure_cnt);
		*len += sprintf(buf + *len, "NACK count                                   %d\n\n", i2c_bus_status_ptr[bus_num].dev_stat[dev_addr].nack_cnt);

			
		}

	//printk (" inside print dev %d  slave %x len %d  buf(addr) %p\n", bus_num, real_addr, *len, (buf+ *len));
	return *len;
}

/*
*********** Device attribute files ***********
*/



static ssize_t 
show_i2c_bus_x_dev(struct kobject *kobj, struct kobj_attribute *attr, char *buf) {

	int bus_num;
	int len = 0;
	int dev_addr;
	int comb_addr;
	int seg_num;
	int mux_slave;
	unsigned char dev_stat_index;
	

#if 1
	/* Get the port number  */
	bus_num = bus_name_to_number(kobj);
	if (bus_num == -EINVAL) {
		
		printk( KERN_INFO "Error : Getting the Port number  \n"); 
		return sprintf(buf, "Error : Getting the Port number  \n");
	}

	/* if not muxed (i.e. dev_48) */
	if (strlen(attr->attr.name) == 6) {
		seg_num=0;
		sscanf(attr->attr.name+4,"%02x",&dev_addr);
	}
	else { /* muxed (i.e. dev_77_3_30) */
		sscanf(attr->attr.name+4,"%02x", &mux_slave);
		sscanf(attr->attr.name+7,"%d", &seg_num);
		sscanf(attr->attr.name+9,"%02x", &dev_addr);
	}
#else	
	sscanf(attr->attr.name+4,"%x",&seg_num);
	sscanf(attr->attr.name+6,"%x",&dev_addr);
#endif	
	
	comb_addr= dev_addr + seg_num;
	
		
	if (i2c_bus_status_ptr[bus_num].hash_table[comb_addr].collision_cnt == 0)
			dev_stat_index = i2c_bus_status_ptr[bus_num].hash_table[comb_addr].dev_stat_index_base;
		else
			dev_stat_index = i2c_bus_status_ptr[bus_num].hash_table[comb_addr].dev_stat_index_base + seg_num;

	
	print_dev_status(bus_num,dev_stat_index, buf, &len);
	
	return len;
	

}


static ssize_t 
clear_i2c_bus_x_dev(struct kobject *kobj, struct kobj_attribute *attr, const char *buf, size_t count) {

/*Note: The following code causes kernel oops. 
 *      Currently, no application clears i2c_bus_x_dev. Disable the following code.
 *      If we want to support clear i2c bus x dev in the future, it needs to fix the problem first. */
#if 0 
	int bus_num;

	int dev_addr;
	int comb_addr;
	int seg_num;
	unsigned char dev_stat_index;
	
	/* Get the port number  */
	bus_num = bus_name_to_number(kobj);
	if (bus_num == -EINVAL) {
		
		printk(KERN_INFO "Error : Getting the Port number  \n"); 
		//return sprintf(buf, "Error : Getting the I2Bus number  \n");
		return -1;
	}

	sscanf(attr->attr.name+4,"%x",&seg_num);
	sscanf(attr->attr.name+6,"%x",&dev_addr);

	
	comb_addr= dev_addr + seg_num;

	
	if (i2c_bus_status_ptr[bus_num].hash_table[comb_addr].collision_cnt == 0)
		dev_stat_index = i2c_bus_status_ptr[bus_num].hash_table[comb_addr].dev_stat_index_base;
	else
		dev_stat_index = i2c_bus_status_ptr[bus_num].hash_table[comb_addr].dev_stat_index_base + seg_num;

	i2c_bus_status_ptr[bus_num].dev_stat[dev_stat_index].mode = 0;
	i2c_bus_status_ptr[bus_num].dev_stat[dev_stat_index].xmit_cnt = 0;
	i2c_bus_status_ptr[bus_num].dev_stat[dev_stat_index].recv_cnt = 0;
	i2c_bus_status_ptr[bus_num].dev_stat[dev_stat_index].recovery_cnt = 0;
	i2c_bus_status_ptr[bus_num].dev_stat[dev_stat_index].busy_cnt = 0;
	i2c_bus_status_ptr[bus_num].dev_stat[dev_stat_index].failure_cnt = 0;
#endif	
	return count;

}


/*
*Attribute Bus group : /sys/kernel/i2c/X Interface ( X = 1 2, 3 Bus)
*/             

static ssize_t 
show_i2c_bus_x_status(struct kobject *kobj, struct kobj_attribute *attr, char *buf) {

	int bus_num;
	int len=0;
	


	/* Get the port number  */
	bus_num = bus_name_to_number(kobj);
	if (bus_num == -EINVAL) {	
		printk( KERN_INFO "Error : Getting the Port number  \n"); 
		return sprintf(buf, "Error : Getting the Port number  \n");
	}

	print_bus_status(bus_num, buf, &len);
	return len;

}


static ssize_t 
clear_i2c_bus_x_status(struct kobject *kobj, struct kobj_attribute *attr, const char *buf, size_t count) {

	int bus_num;
	
	
	/* Get the port number  */
	bus_num = bus_name_to_number(kobj);
	if (bus_num == -EINVAL) {
		printk(KERN_INFO "Error : Getting the Port number  \n"); 
		return -1;
	}


	i2c_bus_status_ptr[bus_num].TX_index = 0;
	i2c_bus_status_ptr[bus_num].MasterRX_index = 0;
	i2c_bus_status_ptr[bus_num].int_cnt = 0;
	i2c_bus_status_ptr[bus_num].bus_recovery_cnt = 0;
	i2c_bus_status_ptr[bus_num].bus_busy_cnt = 0;
	i2c_bus_status_ptr[bus_num].bus_conseq_busy_cnt= 0;

	
	return count;

}


/*
****** Segment attribute files ************
*/


static ssize_t 
show_i2c_bus_2_seg_0(struct kobject *kobj, struct kobj_attribute *attr, char *buf) {

	int len=0;
			
	print_segment_status(0, buf, &len);
	return len;

}
static ssize_t 
show_i2c_bus_2_seg_1(struct kobject *kobj, struct kobj_attribute *attr, char *buf) {

	int len=0;
			
	print_segment_status(1, buf, &len);
	return len;

}
static ssize_t 
show_i2c_bus_2_seg_2(struct kobject *kobj, struct kobj_attribute *attr, char *buf) {

	int len=0;
			
	print_segment_status(2, buf, &len);
	return len;

}
static ssize_t 
show_i2c_bus_2_seg_3(struct kobject *kobj, struct kobj_attribute *attr, char *buf) {

	int len=0;
			
	print_segment_status(3, buf, &len);
	return len;

}
static ssize_t 
show_i2c_bus_2_seg_4(struct kobject *kobj, struct kobj_attribute *attr, char *buf) {

	int len=0;
			
	print_segment_status(4, buf, &len);
	return len;

}

static ssize_t 
show_i2c_bus_2_seg_5(struct kobject *kobj, struct kobj_attribute *attr, char *buf) {

	int len=0;
			
	print_segment_status(5, buf, &len);
	return len;

}
static ssize_t 
show_i2c_bus_2_seg_6(struct kobject *kobj, struct kobj_attribute *attr, char *buf) {

	int len=0;
			
	print_segment_status(6, buf, &len);
	return len;

}
static ssize_t 
show_i2c_bus_2_seg_7(struct kobject *kobj, struct kobj_attribute *attr, char *buf) {

	int len=0;
			
	print_segment_status(7, buf, &len);
	return len;

}
static ssize_t 
clear_i2c_bus_2_seg_0(struct kobject *kobj, struct kobj_attribute *attr, const char *buf, size_t count) {
		
	clear_segment_status(0);
	return count;
}

static ssize_t 
clear_i2c_bus_2_seg_1(struct kobject *kobj, struct kobj_attribute *attr, const char *buf, size_t count) {
		
	clear_segment_status(1);
	return count;
}
static ssize_t 
clear_i2c_bus_2_seg_2(struct kobject *kobj, struct kobj_attribute *attr, const char *buf, size_t count) {
		
	clear_segment_status(2);
	return count;

}
static ssize_t 
clear_i2c_bus_2_seg_3(struct kobject *kobj, struct kobj_attribute *attr, const char *buf, size_t count) {
			
	clear_segment_status(3);
	return count;

}
static ssize_t
clear_i2c_bus_2_seg_4(struct kobject *kobj, struct kobj_attribute *attr, const char *buf, size_t count) {
			
	clear_segment_status(4);
	return count;

}

static ssize_t 
clear_i2c_bus_2_seg_5(struct kobject *kobj, struct kobj_attribute *attr, const char *buf, size_t count) {
			
	clear_segment_status(5);
	return count;

}
static ssize_t 
clear_i2c_bus_2_seg_6(struct kobject *kobj, struct kobj_attribute *attr, const char *buf, size_t count) {
		
	clear_segment_status(6);
	return count;

}
static ssize_t 
clear_i2c_bus_2_seg_7(struct kobject *kobj, struct kobj_attribute *attr, const char *buf, size_t count) {
			
	clear_segment_status(7);
	return count;

}



static struct kobj_attribute i2c_bus_x_status_attribute = 
    __ATTR(status, 0666, show_i2c_bus_x_status, clear_i2c_bus_x_status);
static struct kobj_attribute i2c_bus_2_seg_0_attribute = 
    __ATTR(seg_0, 0666, show_i2c_bus_2_seg_0, clear_i2c_bus_2_seg_0);

static struct kobj_attribute i2c_bus_2_seg_1_attribute = 
    __ATTR(seg_1, 0666, show_i2c_bus_2_seg_1, clear_i2c_bus_2_seg_1);

static struct kobj_attribute i2c_bus_2_seg_2_attribute = 
    __ATTR(seg_2, 0666, show_i2c_bus_2_seg_2, clear_i2c_bus_2_seg_2);

static struct kobj_attribute i2c_bus_2_seg_3_attribute = 
    __ATTR(seg_3, 0666, show_i2c_bus_2_seg_3, clear_i2c_bus_2_seg_3);

static struct kobj_attribute i2c_bus_2_seg_4_attribute = 
    __ATTR(seg_4, 0666, show_i2c_bus_2_seg_4, clear_i2c_bus_2_seg_4);

static struct kobj_attribute i2c_bus_2_seg_5_attribute = 
    __ATTR(seg_5, 0666, show_i2c_bus_2_seg_5, clear_i2c_bus_2_seg_5);

static struct kobj_attribute i2c_bus_2_seg_6_attribute = 
    __ATTR(seg_6, 0666, show_i2c_bus_2_seg_6, clear_i2c_bus_2_seg_6);

static struct kobj_attribute i2c_bus_2_seg_7_attribute = 
    __ATTR(seg_7, 0666, show_i2c_bus_2_seg_7, clear_i2c_bus_2_seg_7);



/*
* Remove /sys/kernel/i2c hierarchy 
*/ 

static void 
vsc_i2c_remove_sysfs(void) {
	int i,j;

#ifdef ARBITRATION_SUPPORT
       del_timer (&i2c_timer);
#endif
	/* We remove all the kobject ( directory) */
	if (i2c_kobj) {

		for (i=0; i < BUS_COUNT; i++) {
		
			if (i2c_ports[i]) {		
				/* Remove the "sys/kernel/i2c/2/seg_x" file  */
				sysfs_remove_file(i2c_ports[i], &i2c_bus_x_status_attribute.attr);
				printk(KERN_INFO "Removed status file (/sys/kernel/i2c/x/status)  file\n");
				if (i==2) {
					sysfs_remove_file(i2c_ports[i], &i2c_bus_2_seg_0_attribute.attr);
					sysfs_remove_file(i2c_ports[i], &i2c_bus_2_seg_1_attribute.attr);
					sysfs_remove_file(i2c_ports[i], &i2c_bus_2_seg_2_attribute.attr);
					sysfs_remove_file(i2c_ports[i], &i2c_bus_2_seg_3_attribute.attr);
					sysfs_remove_file(i2c_ports[i], &i2c_bus_2_seg_4_attribute.attr);
					sysfs_remove_file(i2c_ports[i], &i2c_bus_2_seg_5_attribute.attr);
					sysfs_remove_file(i2c_ports[i], &i2c_bus_2_seg_6_attribute.attr);
					sysfs_remove_file(i2c_ports[i], &i2c_bus_2_seg_7_attribute.attr);
					printk(KERN_INFO "Removed Segment file (/sys/kernel/i2c/2/seg_x)  file\n");
	
				}

				
				for (j=0; j<MAX_NUM_7BIT_SLAVE_ADDR; j++) {
					if (i2c_bus_status_ptr[i].dev_stat[j].present == VALID &&
						(i2c_bus_status_ptr[i].dev_stat[j].counter > 0)) {
						sysfs_remove_file(i2c_ports[i], &i2c_bus_status_ptr[i].dev_stat[j].dev_attribute.attr);
						printk(KERN_INFO "Removed Dev file (/sys/kernel/i2c/%d/dev_%x_0x%x)  file\n",i,
							i2c_bus_status_ptr[i].dev_stat[j].mux_chan,
							i2c_bus_status_ptr[i].dev_stat[j].dev_slave_addr);
						i2c_bus_status_ptr[i].dev_stat[j].counter --;
					}
				}
				
				kobject_put(i2c_ports[i]);  /* void return */
				printk(KERN_INFO "Removed Bus%d \n", i);
			}
		}

		/* Remove the parent object */
		kobject_put(i2c_kobj);  /* void return */
		printk(KERN_INFO "Removed /sys/kernel/i2c hierarchy  ");
	}

}


/*
* Create /sys/kernel/i2c hierarchy 
*/ 


static int 
i2c_create_ports(struct kobject *i2c_kobj) {

	int i; 
	int retval=-ENOMEM;
	const size_t port_len=5;
	char i2c_bus_file_name[port_len];

	i = 0;
		
	do {
		
		snprintf(i2c_bus_file_name, port_len, "%d", i);

		/* Creating the I2C dir */
		i2c_ports[i] = kobject_create_and_add(i2c_bus_file_name, i2c_kobj);
		
		if (i2c_ports[i]) {
			retval = sysfs_create_file(i2c_ports[i], &i2c_bus_x_status_attribute.attr);
			if (retval == 0) {
				printk(KERN_INFO "Created status file (/sys/kernel/i2c/x/status)  file\n");	
			}
			else {
				kobject_put(i2c_ports[i]);  /* void return */
			 	i2c_kobj = NULL;
			 	retval = -ENOMEM;
			}
			if (i == 2) { // for shared bus2, segment attribute files created
				
				retval = sysfs_create_file(i2c_ports[i], &i2c_bus_2_seg_0_attribute.attr);
				if (retval == 0) {
					printk(KERN_INFO "Created segment file (/sys/kernel/i2c/2/seg_0)  file\n");	
				}
				else {
					kobject_put(i2c_ports[i]);  /* void return */
			 		i2c_kobj = NULL;
			 		retval = -ENOMEM;
				}

				
				retval = sysfs_create_file(i2c_ports[i], &i2c_bus_2_seg_1_attribute.attr);
				if (retval == 0) {
					printk(KERN_INFO "Created segment file (/sys/kernel/i2c/2/seg_1)  file\n"); 
				}
				else {
						kobject_put(i2c_ports[i]);	/* void return */
						i2c_kobj = NULL;
						retval = -ENOMEM;
				}

				retval = sysfs_create_file(i2c_ports[i], &i2c_bus_2_seg_2_attribute.attr);
				if (retval == 0) {
					printk(KERN_INFO "Created segment file (/sys/kernel/i2c/2/seg_2)  file\n"); 
				}
				else {
					kobject_put(i2c_ports[i]);	/* void return */
					i2c_kobj = NULL;
					retval = -ENOMEM;
				}

				retval = sysfs_create_file(i2c_ports[i], &i2c_bus_2_seg_3_attribute.attr);
				if (retval == 0) {
					printk(KERN_INFO "Created segment file (/sys/kernel/i2c/2/seg_3)  file\n"); 
				}
				else {
					kobject_put(i2c_ports[i]);	/* void return */
					i2c_kobj = NULL;
					retval = -ENOMEM;
				}
				retval = sysfs_create_file(i2c_ports[i], &i2c_bus_2_seg_4_attribute.attr);
				if (retval == 0) {
					printk(KERN_INFO "Created segment file (/sys/kernel/i2c/2/seg_4)  file\n"); 
				}
				else {
					kobject_put(i2c_ports[i]);	/* void return */
					i2c_kobj = NULL;
					retval = -ENOMEM;
				}
				retval = sysfs_create_file(i2c_ports[i], &i2c_bus_2_seg_5_attribute.attr);
				if (retval == 0) {
					printk(KERN_INFO "Created segment file (/sys/kernel/i2c/2/seg_5)  file\n"); 
				}
				else {
					kobject_put(i2c_ports[i]);	/* void return */
					i2c_kobj = NULL;
					retval = -ENOMEM;
				}
				retval = sysfs_create_file(i2c_ports[i], &i2c_bus_2_seg_6_attribute.attr);
				if (retval == 0) {
					printk(KERN_INFO "Created segment file (/sys/kernel/i2c/2/seg_6)  file\n"); 
				}
				else {
					kobject_put(i2c_ports[i]);	/* void return */
					i2c_kobj = NULL;
					retval = -ENOMEM;
				}
				retval = sysfs_create_file(i2c_ports[i], &i2c_bus_2_seg_7_attribute.attr);
				if (retval == 0) {
					printk(KERN_INFO "Created segment file (/sys/kernel/i2c/2/seg_7)  file\n"); 
				}
				else {
					kobject_put(i2c_ports[i]);	/* void return */
					i2c_kobj = NULL;
					retval = -ENOMEM;
				}
		
			}
		} else {
			printk(KERN_INFO "Error Creating  Bus%s dir \n", i2c_bus_file_name);
			retval = -ENOMEM;
		}

		i++;
			
	} while ((retval == 0) && (i < BUS_COUNT));
		
	return retval;
}

static int 
vsc_i2c_add_sysfs(void) {

	int retval=0;

	i2c_kobj = NULL;
	
	memset(i2c_ports, 0, sizeof(i2c_ports));

	/* extern struct kobject * __must_check kobject_create_and_add(const char *name, struct kobject *parent); 
	*
	* 1) Here, the first argument(name) provides the name of the directory to be created. The second argument (parent 
	* provides the kobject (or directory) under which the directory should be created. 
	*
	* 2) We are going to create a directory under /sys/kernel, hence we have to use 'kernel_kobj' as parent . This function 
	*   returns a pointer to kobject which you can use to create further directories under the directory just created
	*
	*/

	/* Creating the "i2c" dir */
	i2c_kobj = kobject_create_and_add("i2c", kernel_kobj);

	if (i2c_kobj) {

		/*Creating the Bus "0" , "1" ... dir (ie) sys/kernel/i2c/0  */
		retval = i2c_create_ports(i2c_kobj);
		if (retval == 0) {
				printk(KERN_INFO "Created /sys/kernel/i2c hierarchy \n");
				
		}
		else {
			vsc_i2c_remove_sysfs();	
		}
	}
	else {
		retval= -ENOMEM;
	}
	
	return retval;
}





#ifdef ARBITRATION_SUPPORT
#define TIMER_MS 50
static void i2c_timer_function(unsigned long ptr) {
	/* restart short timer */
	RESTART_SHORT_TIMER(RESTART_TIMER_VALUE);	

	mod_timer(&i2c_timer, jiffies + (TIMER_MS*HZ)/1000); //50 ms
}
#endif

static int __init i2c_vsc_init(void) {
	int i=0,j=0;
	int err, ret=0;
	uint32_t i2c_base_addr;

#ifdef ARBITRATION_SUPPORT
	/* init timer */
	init_timer(&i2c_timer);
	i2c_timer.data = (unsigned long)NULL;
	i2c_timer.function = i2c_timer_function;
	i2c_timer.expires = jiffies + ((TIMER_MS*HZ)/1000);
	add_timer (&i2c_timer);
#endif
	printk(KERN_ALERT "Loading VSC452 I2C driver (version %s)\n", VERSION_STR);
	
	g_gpio_dev = get_gpio_dev();

	/* Remap registers */
	i2c_base_addr = I2C_BASE;
	if ((i2c_remap_addr = ioremap_nocache(i2c_base_addr, I2C_REG_SIZE)) == NULL) {
		printk(KERN_ERR "Failed to map I2C IO space %08x-%08x to memory", i2c_base_addr, i2c_base_addr + I2C_REG_SIZE - 1);
		init_iic_status=I2C_REMAP_FAILS;
		ret = -EIO;
		goto fail;
	}

	/* Check I2C memory region */
	if ((err = check_mem_region(i2c_base_addr, I2C_REG_SIZE)) < 0) {
		printk(KERN_ERR "I2C IO space %08x-%08x already in use(err %d)", i2c_base_addr, i2c_base_addr + I2C_REG_SIZE - 1, err);
		init_iic_status=I2C_MEM_CHECK_REG_FAILS;
		ret = -EBUSY;
		goto fail;
	}

	/* Request for the memory region */	
	if (request_mem_region(i2c_base_addr, I2C_REG_SIZE, DRIVER_NAME) == NULL) { 
		printk(KERN_ERR "IO space %08x-%08x i2c request mem region failed (err %d)", i2c_base_addr, i2c_base_addr + I2C_REG_SIZE - 1, err);
		init_iic_status=I2C_REQ_MEM_REG_FAILS;
		ret = -ENODEV;
		goto fail;
	}

	/* Initialize driver's itnernal data structure */
	for (i=0; i<bus_count; i++ ) {
		i2c_init_internal_data(i);
		i2c_init_status_data(i);

	}

	/* 
	 * Registering I2C IRQ for all available bus
	 * I2C driver uses Shared IRQ (IRQ Number 7),shared by all available bus   
 	 */
/*EMW
	 for (i=0; i<bus_count; i++) {
		i2c_vsc_disable_all_interrupts(i);
		if (request_irq(IRQ_I2C, i2c_handler, IRQF_SHARED, "i2c_vsc", &i2c_node[i] ) < 0 ) {
			printk(KERN_ERR "Request for irq %d failed\n", IRQ_I2C);
			init_iic_status=I2C_REGIS_IRQ_FAILS;
			ret = -ENODEV;
			goto fail;
		}
	}
*/
	i2c_vsc_disable_all_interrupts(i);
	if (request_irq(IRQ_I2C, i2c_handler, IRQF_SHARED, "i2c_vsc", &i2c_dev_id ) < 0 ) {
		printk(KERN_ERR "Request for irq %d failed\n", IRQ_I2C);
		init_iic_status=I2C_REGIS_IRQ_FAILS;
		ret = -ENODEV;
		goto fail;
	}

	/* Initialize all I2C bus
	 * By Default all the I2C bus are initialised as Slave ,waiting for the interrupts across the physical hardware bus 
	 */ 
	for (i=0; i<bus_count; i++) {
		i2c_init_hardware(i);
	}

	/* Register ourselves with the upper layers */
	for (i=0; i<bus_count; i++) {
		/* Init bus recovery counter (used for keeping count of bus recovery attempts - informational only) */
		i2c_bus_status_ptr[i].bus_recovery_cnt = 0;
		/* Init bus busy counter */
		i2c_bus_status_ptr[i].bus_busy_cnt = 0;
		i2c_bus_status_ptr[i].bus_conseq_busy_cnt = 0;
		/* Init device failure counter (tracks number of failures per device */
		for (j=0; j<128; j++) {
			i2c_bus_status_ptr[i].dev_stat[j].present = INVALID;
			i2c_bus_status_ptr[i].dev_stat[j].conseq_busy_cnt = 0;
			i2c_bus_status_ptr[i].dev_stat[j].busy_cnt = 0;
			i2c_bus_status_ptr[i].dev_stat[j].conseq_fail_cnt = 0;
			i2c_bus_status_ptr[i].dev_stat[j].failure_cnt= 0;
			i2c_bus_status_ptr[i].dev_stat[j].nack_cnt= 0;
		}

		/* Set up i2c_adapter structs */
		vsc_ops[i].owner = THIS_MODULE;
		vsc_ops[i].class = I2C_CLASS_HWMON;
		vsc_ops[i].algo_data = &vsc_data_ptr[i];
		vsc_ops[i].nr = i;
		sprintf(vsc_ops[i].name, "i2c-vsc452-%d", i);

		/* Add the bus via the algorithm code */
		if (i2c_vsc_add_bus(&vsc_ops[i]) != 0) {
			printk(KERN_ERR "Cannot add bus %d to algorithm layer\n", i);
			return (-ENODEV);
		}
	}
	write_init_st(SET_IIC_OPTION,init_iic_status);
	printk(KERN_ALERT " VSC452 I2C sucessfully loaded i2c statistics REV1\n");



	ret = vsc_i2c_add_sysfs();
	if (ret != 0) {
		goto fail;
	}

	
	return ret;
	
fail:
	i2c_vsc_exit();
	return ret;
}

void i2c_vsc_exit(void) {
	int i;
	uint32_t i2c_base_addr;

	printk (KERN_ALERT "Unloading VSC452 I2C driver ..\n");
	
	if(init_iic_status==0) {
		init_iic_status=0xff;
	}
	write_init_st(SET_IIC_OPTION,0xff);
	/* Remove the bus via algorithm code */
	for (i=0; i<bus_count; i++) {
		i2c_vsc_del_bus(&vsc_ops[i]);
	}

	i2c_base_addr = I2C_BASE;

	/* Free IRQ before unload */
//EMW	for (i=0; i<bus_count; i++) {
//EMW		/* i2c_vsc_disable_all_interrupts(i); */
//EMW		free_irq(IRQ_I2C,&i2c_node[i]);
//EMW	}
	/* i2c_vsc_disable_all_interrupts(i); */
	free_irq(IRQ_I2C,&i2c_dev_id);

	/* Release the memory occupied */  
	release_mem_region(i2c_base_addr,I2C_REG_SIZE);

	/* Unmap registers */
	iounmap(i2c_remap_addr); 

	printk(KERN_ALERT "VSC452 I2C driver Unloaded Successfully\n");
	

	vsc_i2c_remove_sysfs();

	
	return;
}

/* Return bitfield describing our capabilities */
static u32 vsc_func(struct i2c_adapter *adap) {
	return (I2C_FUNC_I2C);
}






int 
set_shared_bus_support_fpga_status( segment_err_T *segment_err) {

	
	if ( segment_err->error_bit & MLER_LOC_REL_SHORT_TMO)
		g_sharedbus_stats[segment_err->segnum].l_rel_short_timeout++;
	if ( segment_err->error_bit & MLER_LOC_REL_LONG_TMO)
		g_sharedbus_stats[segment_err->segnum].l_rel_long_timeout++;
	if ( segment_err->error_bit & MLER_RM_GRT_TMO)
		g_sharedbus_stats[segment_err->segnum].l_rm_grant_timeout++;	
	if ( segment_err->error_bit & MLER_RM_REL_TMO)
		g_sharedbus_stats[segment_err->segnum].l_rm_rel_timeout++;
	if ( segment_err->error_bit & MLER_LOC_SER_EXCH_ERR)
		g_sharedbus_stats[segment_err->segnum].l_serial_exchange_error++;
	if (segment_err->error_bit >> 5 & MRER_RM_REL_SHORT_TMO)
		g_sharedbus_stats[segment_err->segnum].r_rel_short_timeout ++;
	if ( segment_err->error_bit >> 5 & MRER_RM_REL_LONG_TMO)
		g_sharedbus_stats[segment_err->segnum].r_rel_long_timeout ++;	
	if ( segment_err->error_bit >> 5 & MRER_LOC_GRT_TMO)
		g_sharedbus_stats[segment_err->segnum].r_loc_grant_timeout ++;
	if ( segment_err->error_bit >> 5& MRER_LOC_REL_TMO)
		g_sharedbus_stats[segment_err->segnum].r_loc_rel_timeout ++;
	if ( segment_err->error_bit >> 5& MRER_RM_SER_EXCH_ERR)
		g_sharedbus_stats[segment_err->segnum].r_serial_exchange_error ++;
	
	return 0;
	
}



static int
remove_previous_scanned_bus_dev_files(int bus) {
	int j;


	for (j=0; j<MAX_NUM_7BIT_SLAVE_ADDR; j++) {
		if (i2c_bus_status_ptr[bus].dev_stat[j].present == VALID &&
			(i2c_bus_status_ptr[bus].dev_stat[j].counter > 0)) {
			sysfs_remove_file(i2c_ports[bus], &i2c_bus_status_ptr[bus].dev_stat[j].dev_attribute.attr);
			printk(KERN_INFO "Removed Dev file (/sys/kernel/i2c/%d/dev_%x_0x%x)  file\n",bus,
							i2c_bus_status_ptr[bus].dev_stat[j].mux_chan,
							i2c_bus_status_ptr[bus].dev_stat[j].dev_slave_addr);
			i2c_bus_status_ptr[bus].dev_stat[j].counter --;
		}
	}


	return 0;


}


static int 
create_scanned_bus_dev_files(int bus) {



	unsigned char j;
	int retval=-ENOMEM;
 	
	for (j=0; j<MAX_NUM_7BIT_SLAVE_ADDR; j++) {
			
		if (i2c_bus_status_ptr[bus].dev_stat[j].present == VALID) {
					


			if (0 == i2c_bus_status_ptr[bus].dev_stat[j].mux_slave) {
				sprintf(i2c_bus_status_ptr[bus].dev_stat[j].sys_filename, "dev_%02x",
					i2c_bus_status_ptr[bus].dev_stat[j].dev_slave_addr);
			}
			else {

				/* if mux is shared bus (no real mux address so just put "sb") */
				if (0xff == i2c_bus_status_ptr[bus].dev_stat[j].mux_slave) {
					sprintf(i2c_bus_status_ptr[bus].dev_stat[j].sys_filename, "dev_sb_%d_%02x",
						i2c_bus_status_ptr[bus].dev_stat[j].mux_chan, i2c_bus_status_ptr[bus].dev_stat[j].dev_slave_addr);
				}
				else { /* valid mux */
					sprintf(i2c_bus_status_ptr[bus].dev_stat[j].sys_filename, "dev_%02x_%d_%02x",
						i2c_bus_status_ptr[bus].dev_stat[j].mux_slave>>1, i2c_bus_status_ptr[bus].dev_stat[j].mux_chan, i2c_bus_status_ptr[bus].dev_stat[j].dev_slave_addr);
				}
				
			}

			
			/*** search to see if this file has already been created ***/
			if (i2c_bus_status_ptr[bus].dev_stat[j].counter == 0) {
					
				i2c_bus_status_ptr[bus].dev_stat[j].dev_attribute.attr.name = i2c_bus_status_ptr[bus].dev_stat[j].sys_filename;
				i2c_bus_status_ptr[bus].dev_stat[j].dev_attribute.attr.mode = 0666;
				i2c_bus_status_ptr[bus].dev_stat[j].dev_attribute.show = show_i2c_bus_x_dev;
				i2c_bus_status_ptr[bus].dev_stat[j].dev_attribute.store = clear_i2c_bus_x_dev;
				retval = sysfs_create_file(i2c_ports[bus], &i2c_bus_status_ptr[bus].dev_stat[j].dev_attribute.attr);

				if (retval == 0) {
					printk(KERN_INFO "Created Device file (/sys/kernel/i2c/%d/dev_%d_0x%x)  file\n", bus,
						i2c_bus_status_ptr[bus].dev_stat[j].mux_chan,i2c_bus_status_ptr[bus].dev_stat[j].dev_slave_addr); 
					i2c_bus_status_ptr[bus].dev_stat[j].counter ++;
				}
			}
			else {
				printk("Device Files Already Created\n");
			}
 
		}
	}

#if 0
	for (j=0; j<MAX_NUM_7BIT_SLAVE_ADDR; j++) {
		comb_addr = i2c_bus_status_ptr[bus].dev_stat[j].mux_chan + i2c_bus_status_ptr[bus].dev_stat[j].dev_slave_addr;
		printk("222 mux_chan[%d]=%x, slave_addr[%d]=%x,present[%d]=%d, counter[%d]=%ld, c_c=%d, index_base=%x,dup_set=%x\n",
			j,i2c_bus_status_ptr[bus].dev_stat[j].mux_chan,j,i2c_bus_status_ptr[bus].dev_stat[j].dev_slave_addr,
			j,i2c_bus_status_ptr[bus].dev_stat[j].present,j,i2c_bus_status_ptr[bus].dev_stat[j].counter,i2c_bus_status_ptr[bus].hash_table[comb_addr].collision_cnt,
			i2c_bus_status_ptr[bus].hash_table[comb_addr].dev_stat_index_base,i2c_bus_status_ptr[bus].duplicate_set_cnt);
	}
#endif
			
	return 0;
}

static int
clear_hash_table( int bus) {
	int i;

	i2c_bus_status_ptr[bus].dev_index_cnt = 0;
	i2c_bus_status_ptr[bus].duplicate_set_cnt = 0;
	
	for (i=0; i<MAX_COMB_ADDR; i++) {
		i2c_bus_status_ptr[bus].hash_table[i].collision_cnt = HASH_INDEX_VACANT;
		i2c_bus_status_ptr[bus].hash_table[i].dev_stat_index_base = 0x0;

	}


	for (i=0;i<MAX_NUM_7BIT_SLAVE_ADDR;i++) {
		//i2c_bus_status_ptr[bus].dev_stat[i].counter = 0;
		i2c_bus_status_ptr[bus].dev_stat[i].present = 0;	
	}

	return 0;
}



static int
set_dev_info(i2c_info_t i2c_info) {


	unsigned char comb_addr=0;
	unsigned char dev_stat_index=0;
	int tmp_collision_cnt;
	unsigned char tmp_duplicate_set_cnt;
	unsigned char tmp_seg;
	unsigned char new_dev_index;
	//unsigned char j;

	
	comb_addr = i2c_info.dev_slave +  i2c_info.mux_chan;
	
	// We first get the collison_cnt from the hash table , which shall be -1 by default
	tmp_collision_cnt = i2c_bus_status_ptr[i2c_info.bus_num].hash_table[comb_addr].collision_cnt;


	if (tmp_collision_cnt == HASH_INDEX_VACANT) { // This slot at the index of comb_addr is vacant.


		// Store the dev_stat_index_base as dev_index_cnt
		i2c_bus_status_ptr[i2c_info.bus_num].hash_table[comb_addr].dev_stat_index_base = i2c_bus_status_ptr[i2c_info.bus_num].dev_index_cnt;

		// Increment the dev_index_cnt 
		i2c_bus_status_ptr[i2c_info.bus_num].dev_index_cnt++;

		i2c_bus_status_ptr[i2c_info.bus_num].hash_table[comb_addr].collision_cnt = HASH_INDEX_OCCUPIED;

		dev_stat_index = i2c_bus_status_ptr[i2c_info.bus_num].hash_table[comb_addr].dev_stat_index_base;
		i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[dev_stat_index].present = VALID;
		i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[dev_stat_index].mux_slave  = (unsigned char)i2c_info.mux_slave;
		i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[dev_stat_index].mux_chan	 = (unsigned char)i2c_info.mux_chan;
		i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[dev_stat_index].dev_slave_addr = (unsigned char)i2c_info.dev_slave;
		
	}
	else {  // We have a collison 

		// First time collison 
	
		/* We shall copy the "no collison index" to the colloison index space */
		if (tmp_collision_cnt == HASH_INDEX_OCCUPIED) {
			
			// First we read the [dev_stat_index_base] stored in HASH table for for this index   
			dev_stat_index = i2c_bus_status_ptr[i2c_info.bus_num].hash_table[comb_addr].dev_stat_index_base;

			// Get the tmp segment
			tmp_seg = i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[dev_stat_index].mux_chan;

			tmp_duplicate_set_cnt = i2c_bus_status_ptr[i2c_info.bus_num].duplicate_set_cnt;
					 

			new_dev_index =  MAX_NON_COLLISION_DEV + (tmp_duplicate_set_cnt * SEGMENT_COUNT) + tmp_seg ;
				 
	   		// Move the previous store value from orderly index to the collision set (dev_index > 64)
	   				 	
			i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[new_dev_index].present = VALID;
			i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[new_dev_index].mux_slave  = i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[dev_stat_index].mux_slave;
			i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[new_dev_index].mux_chan	 = i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[dev_stat_index].mux_chan;
			i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[new_dev_index].dev_slave_addr = i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[dev_stat_index].dev_slave_addr;

			// no collision index is not present or valid for dev_stat value
			i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[dev_stat_index].present = 0;
			
			// New value for collisioned comb addr

			new_dev_index =  MAX_NON_COLLISION_DEV + (tmp_duplicate_set_cnt * SEGMENT_COUNT) + i2c_info.mux_chan ;
			i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[new_dev_index].present = VALID;
			i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[new_dev_index].mux_slave  = (unsigned char)i2c_info.mux_slave;
			i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[new_dev_index].mux_chan	 = (unsigned char)i2c_info.mux_chan;
			i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[new_dev_index].dev_slave_addr = (unsigned char)i2c_info.dev_slave;
											
			// Update the has table with the base index and colliction cnt

			// Store the dev_stat_index_base as dev_index_cnt
			i2c_bus_status_ptr[i2c_info.bus_num].hash_table[comb_addr].dev_stat_index_base = MAX_NON_COLLISION_DEV +(tmp_duplicate_set_cnt * SEGMENT_COUNT);

			i2c_bus_status_ptr[i2c_info.bus_num].hash_table[comb_addr].collision_cnt++;

						
			i2c_bus_status_ptr[i2c_info.bus_num].duplicate_set_cnt++;
 
		}
		else  { // Second and latter dupication

			new_dev_index =  i2c_bus_status_ptr[i2c_info.bus_num].hash_table[comb_addr].dev_stat_index_base + i2c_info.mux_chan;
			i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[new_dev_index].present = VALID;
			i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[new_dev_index].mux_slave  = (unsigned char)i2c_info.mux_slave;
			i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[new_dev_index].mux_chan	 = (unsigned char)i2c_info.mux_chan;
			i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[new_dev_index].dev_slave_addr = (unsigned char)i2c_info.dev_slave;

					
			i2c_bus_status_ptr[i2c_info.bus_num].hash_table[comb_addr].collision_cnt ++;
					
					
		}
	
	}

	


#if 0
	for (j=0; j<MAX_NUM_7BIT_SLAVE_ADDR; j++) {
		comb_addr = i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[j].mux_chan + i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[j].dev_slave_addr;
		printk("111 mux_chan[%d]=%x, slave_addr[%d]=%x,present[%d]=%d, counter[%d]=%ld, c_c=%d, index_base=%x,dup_set=%x\n",
			j,i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[j].mux_chan,j,i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[j].dev_slave_addr,
			j,i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[j].present,j,i2c_bus_status_ptr[i2c_info.bus_num].dev_stat[j].counter,
			i2c_bus_status_ptr[i2c_info.bus_num].hash_table[comb_addr].collision_cnt,
			i2c_bus_status_ptr[i2c_info.bus_num].hash_table[comb_addr].dev_stat_index_base,i2c_bus_status_ptr[i2c_info.bus_num].duplicate_set_cnt);
	}

#endif
	return 0;
}


static int
add_dev_file(i2c_info_t i2c_info) {

	unsigned char comb_addr=0;
	int bus = i2c_info.bus_num;
	int retval=-ENOMEM;
	unsigned char dev_stat_index;

	

	set_dev_info(i2c_info);
	
	comb_addr = i2c_info.dev_slave +  i2c_info.mux_chan;

	if (i2c_bus_status_ptr[i2c_info.bus_num].hash_table[comb_addr].collision_cnt == 0)
		dev_stat_index = i2c_bus_status_ptr[i2c_info.bus_num].hash_table[comb_addr].dev_stat_index_base;
	else 
		dev_stat_index = i2c_bus_status_ptr[i2c_info.bus_num].hash_table[comb_addr].dev_stat_index_base + i2c_info.mux_chan;

	

	/** if this file haven't created before */
	if (i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].counter == 0){

		sprintf(i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].sys_filename, "dev_%d_0x%x",
					i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].mux_chan, i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].dev_slave_addr);
		
		i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].dev_attribute.attr.name = i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].sys_filename;
		i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].dev_attribute.attr.mode = 0666;
		i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].dev_attribute.show = show_i2c_bus_x_dev;
		i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].dev_attribute.store = clear_i2c_bus_x_dev;
		retval = sysfs_create_file(i2c_ports[bus], &i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].dev_attribute.attr);

		if (retval == 0) {
			printk(KERN_INFO "Add Device file (/sys/kernel/i2c/%d/dev_%d_0x%x)  file\n", bus,
							i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].mux_chan,i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].dev_slave_addr); 
			i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].counter ++;
		}
	}
	else {
		printk("Device Files Already Created\n");
	}

	return 0;
}


static int
delete_dev_file(i2c_info_t i2c_info) {
	unsigned char comb_addr=0;
	int bus=i2c_info.bus_num;
	unsigned char dev_stat_index;

	
	comb_addr = i2c_info.dev_slave +  i2c_info.mux_chan;

	if (i2c_bus_status_ptr[i2c_info.bus_num].hash_table[comb_addr].collision_cnt == 0)
		dev_stat_index = i2c_bus_status_ptr[i2c_info.bus_num].hash_table[comb_addr].dev_stat_index_base;
	else 
		dev_stat_index = i2c_bus_status_ptr[i2c_info.bus_num].hash_table[comb_addr].dev_stat_index_base + i2c_info.mux_chan;
	

	/*** only delete this dev file if it is created already */
	if ((i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].present == VALID) && 
		(i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].counter > 0)) {
		sysfs_remove_file(i2c_ports[bus],&i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].dev_attribute.attr);
			printk(KERN_INFO "Removed Dev file (/sys/kernel/i2c/%d/dev_%x_0x%x)  file\n",bus,i2c_info.mux_chan,
				i2c_info.dev_slave);
		i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].counter --;	
	}
	else
		printk("Cann't delet no exist file \n"); 
	return 0;

}




/* Algorithm level handling of ioctls and other control commands */
static int vsc_control(struct i2c_adapter *i2c_adap, 
			unsigned long cmd, unsigned long arg) {
	int retval = 0;
	bus_recovery_info_T  bus_recovery_info;

	segment_err_T segment_err_bit;
	i2c_info_t i2c_info;
	int bus;
	memset(&i2c_info, 0, sizeof(i2c_info_t));


	switch (cmd) {
#ifdef I2C_EXTENDED_CMDS
	case I2C_SLAVEREAD:
		/* Unsupported (for now) */
		retval = -ENOTTY;
		break;

	case I2C_SLAVEWRITE:
		/* Unsupported (for now) */
		retval = -ENOTTY;
		break;

	case I2C_SET_HOST_ADDR:
		/* i2c_vsc_set_slave( i2c_adap->nr, arg ); */
		/* Unsupported (for now) */
		retval = -ENOTTY;
		break;

	case I2C_GET_HOST_ADDR:
            	/* retval = i2c_vsc_get_slave( i2c_adap->nr ); */
		/* Unsupported (for now) */
		retval = -ENOTTY;
		break;

	case I2C_RESET:
		printk(KERN_ALERT "I2C%d: Resetting the controller\n", i2c_adap->nr);
		/* Reset the controller */
		i2c_init_internal_data(i2c_adap->nr);
		i2c_init_hardware(i2c_adap->nr);
		break;

	case I2C_RECOVER:
		if(0 != i2c_bus_recovery(i2c_adap->nr)){
			return -EFAULT;
		}
		break; 

#endif /* I2C_EXTENDED_CMDS */
	case I2C_SET_REC_INFO:
		if (copy_from_user(&bus_recovery_info, (void*)arg, sizeof(bus_recovery_info_T)))
			return -EFAULT;

		if (0 != vsc_set_recovery_info(i2c_adap->nr, &bus_recovery_info)) {
			dev_err(&i2c_adap->dev, "vsc_control: set recovery info vsc\n");
		}
		break;

	case I2C_GET_REC_INFO:
		if (0 != vsc_get_recovery_info(i2c_adap->nr, &bus_recovery_info)) {
			dev_err(&i2c_adap->dev, "vsc_control: get recovery info vsc\n");
		}

		if (copy_to_user((void*)arg, &bus_recovery_info, sizeof(bus_recovery_info_T))) 
			return -EFAULT;

		break;


	case PASS_ERROR_BIT_TO_I2C_DRV:
		if (copy_from_user (&segment_err_bit, (void*)arg, sizeof(segment_err_T)))
			return -EFAULT;
		if ( 0 != set_shared_bus_support_fpga_status(&segment_err_bit)) {
			dev_err(&i2c_adap->dev, "vsc_control: set support fpga fail\n");
		}
		break;


	case SCAN_DEVICE:
		if (copy_from_user (&i2c_info, (void*)arg, sizeof(i2c_info_t)))
			return -EFAULT;
		if (0 != set_dev_info(i2c_info)) {
			dev_err(&i2c_adap->dev, "vsc_control: scan_device fail\n");
		}
		break;



	case CREATE_SCANNED_DEV_SYS_FILES:
		if (copy_from_user (&bus, (void*)arg, sizeof(int)))
			return -EFAULT;

		if (0 != create_scanned_bus_dev_files(bus)) {
			dev_err(&i2c_adap->dev, "vsc_control: set mux slave and mux chan fail\n");
		}
		
		break;

	case ADD_DEV_SYS_FILE:
		if (copy_from_user (&i2c_info, (void*)arg, sizeof(i2c_info_t)))
					return -EFAULT; 
		if (0 != add_dev_file(i2c_info)) {
			dev_err(&i2c_adap->dev, "vsc_control: scan_device fail\n");
		}
		
		break;

	case DELET_DEV_SYS_FILE:
		if (copy_from_user (&i2c_info, (void*)arg, sizeof(i2c_info_t)))
					return -EFAULT;
		if (0 != delete_dev_file(i2c_info)) {
			dev_err(&i2c_adap->dev, "vsc_control: scan_device fail\n");
		}
		break;

	case CLEAR_HASH_TABLE:
			if (copy_from_user (&bus, (void*)arg, sizeof(int)))
				return -EFAULT;
	
			if (0 != clear_hash_table(bus)) {
				dev_err(&i2c_adap->dev, "vsc_control: scan_device fail\n");
			}
			break;

	case REMOVE_PREVIOUS_SCANNED_DEV_SYS_FILES:
		if (copy_from_user (&bus, (void*)arg, sizeof(int)))
			return -EFAULT;

		if (0 != remove_previous_scanned_bus_dev_files(bus)) {
			dev_err(&i2c_adap->dev, "vsc_control: set mux slave and mux chan fail\n");
		}
		
		break;
	
	default:
		dev_err(&i2c_adap->dev, "algo_control: Unknown command\n");
		retval = -ENOTTY;
	}

	return retval;
}

static struct i2c_algorithm vsc_algo = {
	.master_xfer = vsc_master_xfer,
	.functionality = vsc_func,
	.i2c_hw_ioctl = vsc_control,
	.slave_recv = vsc_slave_xfer,
#ifdef I2C_BUS_RECOVERY
	.i2c_recovery = vsc_i2c_bus_recovery,
#else
	.i2c_recovery = NULL,
#endif
};


int 
i2c_vsc_add_bus( struct i2c_adapter *i2c_adap ) {
    /* Register new adapter */
    i2c_adap->algo = &vsc_algo;

    /* Set some sensible defaults */
    i2c_adap->timeout = DEFAULT_TIMEOUT;
    i2c_adap->retries = DEFAULT_RETRIES;

    i2c_add_adapter( i2c_adap );
    return(0);
}

int 
i2c_vsc_del_bus( struct i2c_adapter *i2c_adap ) {
    return( i2c_del_adapter( i2c_adap ) );
}


u8
i2c_vsc_get_dev_stat_index(int bus, u8 slave_addr, u8 seg_num) {

	u8 dev_stat_index = 0;
	int combo_addr = (slave_addr + seg_num);
	
	if (i2c_bus_status_ptr[bus].hash_table[combo_addr].collision_cnt == 0)
		dev_stat_index = (u8)i2c_bus_status_ptr[bus].hash_table[combo_addr].dev_stat_index_base;
	else
		dev_stat_index = (u8)(i2c_bus_status_ptr[bus].hash_table[combo_addr].dev_stat_index_base + seg_num);

	/* Make sure device is VALID (listed in I2C map file) */
	if (i2c_bus_status_ptr[bus].dev_stat[dev_stat_index].present == VALID) {
		return dev_stat_index;
	}
	else { /* Device is unknown to us (no slave address defined in I2C map file) */

		/* Use empty location */
		return MAX_DEV_STAT_INDEX; 
	}
}

MODULE_DESCRIPTION("This is VSC452 I2C driver module.");
MODULE_LICENSE("GPL");

module_init( i2c_vsc_init );
module_exit( i2c_vsc_exit );
